// $Id: messagesocket.cpp,v 1.2 2000/08/14 05:09:30 garbanzo Exp $

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/types.h>
#include <sys/socket.h>

#include <netinet/in.h>
#include <arpa/inet.h>

#ifdef HAVE_SYS_SOCKIO_H
#include <sys/sockio.h>
#endif /* HAVE_SYS_SOCKIO_H */

#include <netdb.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <messagesocket.h>


MessageSocket::MessageSocket( void )
{
	type = None;
	socketfd=-1;
	bzero(&ip_addr.in6_addr,16);
}


MessageSocket::~MessageSocket( void )
{
}


void MessageSocket::disableIpv6Lookups() {
	ipv6_lookups_enabled=0;
}


InetType MessageSocket::intype=InetNone;
int      MessageSocket::ipv6_lookups_enabled=1;


// returns 0: failure
//         4: remote address is IPv4
//         6: remote address is IPv6
int MessageSocket::lookupHostname( const char *hostname, unsigned char *store, const int len) {
	struct hostent *he=NULL;
	unsigned long *p;
	unsigned short *s;
	char b[100];

	// rip brackets from IPv6 reference
	if (*hostname=='[') {
		hostname++;
		char *cb=strchr(hostname,']');
		if (!cb) return 0;
		*cb=0;
	}
	printf ("MessageSocket::lookupHostname(): resolving Hostname %s\n",hostname);
	if (intype==Inet6) {
		if ( ipv6_lookups_enabled ) { 
			he = gethostbyname2 (hostname, AF_INET6);
		}
		if (he) { if (len>=16) { bcopy (he->h_addr,store,16); goto ok6; } else goto fail0; }
		else {
			printf( "MessageSocket::lookupHostname(): gethostbyname2(AF_INET6) failed%s.\n", ipv6_lookups_enabled?"":" (disabled) " );
			he = gethostbyname2 (hostname, AF_INET);
			if ( !he ) {
				printf( "MessageSocket::lookupHostname(): gethostbyname2(AF_INET) failed too.\n" );
				goto fail0;
			}
			if (len<16) goto fail0;
			// construct mapped address
			p=(unsigned long *)(store);
			p[0]=0;
			p[1]=0;
			p[2]=0;  s=((unsigned short *)p)+5;  *s=0xffff; 
			p[3]=*((unsigned long*)he->h_addr);
			printf ("MessageSocket::lookupHostname(): ok -> %s\n",inet_ntop(AF_INET6, store, b, 100));
			return 4;
		}
	} else {
		he = gethostbyname2 (hostname, AF_INET);
		if ( !he ) {
			printf( "MessageSocket::lookupHostname(): gethostbyname2(AF_INET) failed.\n" );
			goto fail0;
		}
		if (len>=4) bcopy(he->h_addr,store,4); else goto fail0;
	}

	printf ("MessageSocket::lookupHostname(): ok -> %s\n",inet_ntop(AF_INET, store, b, 100));
	return 4;
ok6:	
	printf ("MessageSocket::lookupHostname(): ok -> %s\n",inet_ntop(AF_INET6, store, b, 100));
	return 6;
fail0:
	printf ("MessageSocket::lookupHostname(): failed\n");
	return 0;
}


// returns 0: failure
//         4: remote address is IPv4
//         6: remote address is IPv6
int MessageSocket::getHostIpType( const char *hostname) {
	unsigned char b[16];
	return lookupHostname( hostname, b, 16);
}


// returns 0: failure
//         4: remote address is IPv4
//         6: remote address is IPv6
int MessageSocket::setHostname( const char *hostname )
{
	return lookupHostname( hostname, (unsigned char *)(ip_addr.raw_addr), 16);
}


int MessageSocket::getSocketIpVersion() {
	if (intype==Inet6) return 6;
	if (intype==Inet4) return 4;
	return 0;
}


void MessageSocket::forcePortNumber( unsigned int newport )
{
	ourport = newport;
}


// returns 4: numeric IPv4
//         6: numeric IPv6
//         0: none of both, probalby a hostname
int MessageSocket::getAddressVersion( const char *addr ) {
	char b[100];

	if ( inet_pton(AF_INET, addr, b) == 0) { // not IPv4
		if ( inet_pton( AF_INET6, addr, b ) == 0 ) // not IPv6
			return 0;
		else 
			return 6;
	} else {
		return 4;
	}
}


