#include <sys/types.h>
#include <sys/socket.h>
#include <strings.h>
#include <unistd.h>
#include <stdio.h>
#include <tcpmessagesocket.h>

TCPMessageSocket::TCPMessageSocket( void )
{
	type = SocketTCP;

	if ( ( socketfd = socket( AF_INET, SOCK_STREAM, 0 ) ) == -1 ) {
		perror( "TCPMessageSocket(): socket() failed" );
	}
}

TCPMessageSocket::TCPMessageSocket( int newfd )
{
	type = SocketTCP;

	socketfd = newfd;
}

TCPMessageSocket::~TCPMessageSocket( void )
{
	close( socketfd );
}

int TCPMessageSocket::connect( unsigned int portnum )
{
	socketaddress.soa4.sin_family = AF_INET;
	socketaddress.soa4.sin_port = htons( portnum );
	socketaddress.soa4.sin_addr = ip_addr.in4_addr;
	bzero( &( socketaddress.soa4.sin_zero ), 8 );

	if ( ::connect( socketfd, &socketaddress.soa, sizeof( struct sockaddr ) ) == -1 ) {
		perror( "TCPMessageSocket::connect(): connect() failed" );
		return -1;
	}

	return 0;
}

int TCPMessageSocket::send( const char *sendbuffer, unsigned int length )
{
//	if ( ::send( socketfd, sendbuffer, strlen( sendbuffer ), 0 ) == -1 ) {
	if ( ::send( socketfd, sendbuffer, length, 0 ) == -1 ) {
		perror( "TCPMessageSocket::send(): send() failed" );
		return -1;
	}

	return 0;
}

int TCPMessageSocket::receive( char *recvbuffer, unsigned int maxlength )
{
	int numbytes;

	if ( ( numbytes = recv( socketfd, recvbuffer, maxlength, 0 ) ) == -1 ) {
		perror( "TCPMessageSocket::receive(): recv() failed" );
		return -1;
	}

	return numbytes;
}

unsigned int TCPMessageSocket::listen( unsigned int portnum )
{
	socketaddress.soa4.sin_family = AF_INET;
	socketaddress.soa4.sin_port = htons( portnum );
	socketaddress.soa4.sin_addr.s_addr = INADDR_ANY;
	bzero( &( socketaddress.soa4.sin_zero ), 8 ); // is this portable?

	int count = 0;
	while ( ::bind( socketfd, &socketaddress.soa, sizeof( struct sockaddr ) ) == -1  && count <= 10 ) {
		count++;
		portnum += 2;
		socketaddress.soa4.sin_port = htons( portnum );
	}
	if ( count > 10 ) {
		perror( "TCPMessageSocket::listen(): bind() failed" );
		return 0;
	}
	if ( ::listen( socketfd, 10 ) == -1 ) {
		perror( "TCPMessageSocket::listen(): listen() failed" );
		return 0;
	}

	return portnum;
}

int TCPMessageSocket::accept( void )
{
	int connectfd;
	socklen_t sockaddr_in_size;

	sockaddr_in_size = sizeof( struct sockaddr_in );

	if ( ( connectfd = ::accept( socketfd, &socketaddress.soa, &sockaddr_in_size ) ) == -1 ) {
		perror( "TCPMessageSocket::accept(): accept() failed" );
		return -1;
	}

	return connectfd;
}

int TCPMessageSocket::listenOnEvenPort( void )
{
	return -1;
}

